/* Parallel Trapezoidal Rule
 *
 * Input: None.
 * Output:  Estimate of the integral from 0 to Pi/2 of
 *              f(x) = dx/(sqrt(1-(cos x * cos x)/2))
 *          using the trapezoidal rule and 128 trapezoids.
 *
 * Algorithm:
 *    1.  Each process calculates "its" interval of
 *        integration.
 *    2.  Each process estimates the integral of f(x)
 *        over its interval using the trapezoidal rule.
 *    3a. Each process != 0 sends its integral to 0.
 *    3b. Process 0 sums the calculations received from
 *        the individual processes and prints the result.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* using MPI routines, definitions, etc. */
#include "mpi.h"


main(int argc, char** argv) {
    int         each_process_rank;   /* Each process rank */
    int         process_num;         /* The number of processes */
    double      pi = 3.141592654; /* the value of pi */
    double      range_a = 0.0;   /* Left endpoint */
    double      range_b = 0.5*pi;   /* Right endpoint */
    int         trapezoid_num = 128;  /* Number of trapezoids */
    double      trapezoid_b_len;  /* Trapezoid base length */
    double      local_range_a;   /* Left endpoint each process */
    int         local_trapezoid_num; /* Number of trapezoids for */
                                     /* my calculation */
    double      integral;  /* Integral over my interval */
    double      total_integral;  /* Total integral */
    int         process_source;    /* Process sending integral  */
    int         process_dest = 0;  /* All messages go to Process 0 */
    int         tag = 0;
    MPI_Status  status;

    double trapezoid_cal(double a, int n, 
                  double h);  /* Calculate local integral  */

    /* Let the system do what it needs to start up MPI */
    MPI_Init(&argc, &argv);

    /* Get each process rank */
    MPI_Comm_rank(MPI_COMM_WORLD, &each_process_rank);

    /* Find out how many processes are being used */
    MPI_Comm_size(MPI_COMM_WORLD, &process_num);

    /* The base of the trapezoid is the same for all processes */
    trapezoid_b_len = (range_b-range_a)/(double)trapezoid_num;

    /* It is the number of trapezoids */
    local_trapezoid_num = trapezoid_num/process_num;

    /* Length of each process' interval of
     * integration = local_trapezoid_num*trapezoid_b_len.  So my interval
     * starts at: */
    local_range_a = range_a + (double)each_process_rank*(double)local_trapezoid_num*trapezoid_b_len;

    integral = trapezoid_cal(local_range_a, local_trapezoid_num, trapezoid_b_len);

    /* Add up the integrals calculated by each process */
    if (each_process_rank == 0) {
	printf("The Contribution of Process %d is %f\n", each_process_rank,integral);
        total_integral = integral;
        for (process_source = 1; process_source < process_num; process_source++) {
            MPI_Recv(&integral, 1, MPI_DOUBLE, process_source, tag,
                MPI_COMM_WORLD, &status);
            printf("The Contribution of Process %d is %f\n", process_source, integral);
            total_integral = total_integral + integral;
        }
    } else {  
        MPI_Send(&integral, 1, MPI_DOUBLE, process_dest,
            tag, MPI_COMM_WORLD);
    }

    /* Print the result */
    if (each_process_rank == 0) {
        printf("With the number of trapezoid = %d trapezoids, our estimate\n",
            trapezoid_num);
        printf("of the integral from %f to %f = %f\n",
            range_a, range_b, total_integral);
    }

    /* Shut down MPI */
    MPI_Finalize();
} /*  main  */


/*
 * each processor calculates the part of the trapezoid,
 * 0.5*h*(f(x0) + f(x1)) + ......+ 0.5*h*(f(xi-1) + f(xi))
 */

double trapezoid_cal( 
          double local_range_a /* in */,
          int local_n   /* in */,
          double h         /* in */) 
{
    double integral = 0.0; /* Store result in integral  */
    double right_x; /* the right end point value */
    double f_left_x;  /* the function valuse of the left side */
    double f_right_x;  /* the function value of the right end point */
    int i; /* for statement i */

    double f(double x); /* function we're integrating */

    /* Calculate the start point of the trapezoid */
    f_left_x = f(local_range_a);

    /* Store the increment on the right end point */
    right_x = local_range_a;

    /* each process calculate the number of the part of the trapezoid */
    for (i = 1; i <= local_n; i++) {

	/* Increment the base of the trapezoid on the right end point */ 
        right_x = right_x + h;

	/* Calculate the right end point value */
        f_right_x = f(right_x);

	/* Accumulate the trapezoidal area */
        integral += 0.5*h*(f_left_x + f_right_x);

	/* Indicate the next left end point */
        f_left_x = f_right_x;
    }  /* for statement */

    return integral;  /* the final value return */

} /* Trapezoid Calculation */


/*
 * the definition of the function f(x)
 */

double f(double x) {

    double cos_val;  /* Temparary value for cosine function */

    cos_val = cos(x); /* Cosine function */

    /* Store calculation in return_val. */
    return (1.0/sqrt(1.0 - 0.5*cos_val*cos_val));
} /* f */
